%% This Version: May 18, 2021. @copyright Rui Da and Dacheng Xiu
% If you use these codes, please cite our paper 
% "When Moving-Average Models Meet High-Frequency Data: Uniform Inference on Volatility." (2021)

% This code is used for both verification of CLT in simulations and
% comparsion with alternative estimators:  RV, RV (5min), MA(1), PVG, and FRK. 
% Three QMLE models are considered: AIC, AICstar (BIC), and quse = 0 (i.e., MA(1)).
% Uniformly valid confidence intervals are also constructed.

%%

function Simulation_Final(ii)

tic;

%%  Initialization

J          =    5;    % number of lags of autocorrelations to calculate for PVG and FRK
M          =    1000; % number of Monte Carlos
T          =    1/252;
nbseconds  =    5;
sublag     =    300/nbseconds;  % 5 min subsampling frequency
deltabar   =    T/23400*nbseconds;

dist       =    1;

switch dist
    case 0 % Standardized t distribution
        nu  = 7;
        cum = 6/(nu-4);
    case 1 % Standardized normal    
        cum = 0;
    case 2 % Standardized Uniform
        cum = (2*sqrt(3))^4/4*(-1/30);
end


for     zz = [0.1, 0.5, 2.5] % small, median, large noise magnitudes.


    iota2true  =    (zz/1000)^2;
    thetatrue  =    0.2;
    q          =    length(thetatrue);
    phitrue    =    0.4;
    p          =    length(phitrue);
    quse       =    0;  %Compare with i.i.d. Noise.
    qtrue      =    round(log(T/deltabar)); % alpha in the paper

    gammatrue  =    real(iota2true/2/pi*integral(@(lambda)exp(1i.*lambda.*(0:J)).*abs(1 + sum(exp(1i*lambda*(1:q)').*thetatrue'))^2./abs(1-sum(exp(1i*lambda*(1:p)').*phitrue')).^2 ,-pi,pi,'ArrayValued',true)); %*sum(thetatrue(k:J).*[1,thetatrue(1:J-k)]);

    alpha      =    -2.8;
    beta       =    6;
    kappa      =    -4;
    rho        =    -0.8;
    sigma      =    0.8; 

    X0         =    log(40);
    F0         =    0;

    mu0        =    0.05;
    mu1        =    0.5;
    lambda     =    25;
    muJ        =    0.02;
    sigmaJ     =    0.02;
    mux        =    0;
    sigmax     =    0.05;

    if iota2true>0
        model             =     arima('Constant',0,'AR',phitrue,'MA',thetatrue,'Variance',1); 
    else
        model = [];
    end
    
    % Tuning parameters for JLZ(PVG) and FRK
    
    knlist = [2,4,6];
    hnlist = [0.6,0.8,1.0];
    knplist    = [3,6,9];
    kntemp    = repmat(knlist,1,length(hnlist));
    hntemp    = kron(hnlist,ones(1,length(knlist))); 
    knl = repmat(kntemp,1,length(knplist));
    hnl = repmat(hntemp,1,length(knplist));
    knpl= kron(knplist,ones(1,length(kntemp)));
       
    rknlist = [0.5,1,1.5];
    jknlist = [0, 5, 10];
    hknlist = [25,50,75];
        
    rkntemp    = repmat(rknlist,1,length(jknlist));
    jkntemp    = kron(jknlist,ones(1,length(rknlist))); 
    rkn = repmat(rkntemp,1,length(hknlist));
    jkn = repmat(jkntemp,1,length(hknlist));
    hkn = kron(hknlist,ones(1,length(rkntemp)));
    
%------ Initialization -------------
    RV_hat            =     zeros(M,1);
    RVsub_hat         =     zeros(M,1);
    RV_true           =     zeros(M,1);    
    QML_var           =     zeros(M,3);   
    QML_iota   = zeros(M,3);
   
    QML_fval   = zeros(M,3);
    QML_BIC    = zeros(M,1);
    QML_AIC    = zeros(M,1);
    QML_qhat   = zeros(M,3);

    QML_std    = zeros(M,3);
    QML_sstd   = zeros(M,3);
    QML_ustd   = zeros(M,3);
    QML_ma     = nan(12,3,M);
    QML_gamma  = nan(13,3,M);
    QML_rho    = nan(12,3,M);

    NUM        = zeros(M,1);

    PVG_var    = zeros(M,length(knl));      
    PVG_gamma  = nan(J+1,length(knl),M);
    PVG_rho    = nan(J,length(knl),M);
    PVG_gamma_std = nan(J+1,length(knl),M);
    PVG_rho_std    = nan(J,length(knl),M);
    RK_var = zeros(M,length(rkn));   


for iMC        = ((ii-1)*10+1):(ii*10)

%parfor iMC        = 1:M     % Monte Carlo

  
    rng(iMC); 
    ticID      =     tic;
       
    %possion arrival
    
    rate       =     @(t) 1+cos(t/T*2*pi)/2;
    n  = round(T/deltabar);
    De_t_0     =     exprnd(T/n,2*n,1);
    De_t       =     zeros(2*n,1);
    De_t(1)    =     De_t_0(1)/rate(0);
    xi         =     zeros(2*n,1);
    for i      =     2:2*n
        De_t(i)=     De_t_0(i)/rate(sum(De_t(1:i-1)));
        xi(i-1)=     1/rate(sum(De_t(1:i-1)));
    end

    time       =     cumsum(De_t);

    if (sum(De_t)>=T)
        N      =     find(time>=T,1);
        time   =     time(1:N);
        De_t   =     De_t(1:N);
    end
    xi         =     xi(1:N);
    time       =     [0;time];
    time       =     time/time(end);

    %  noise

    mu_G       =    1+0.1*cos(2*pi*time);  
    sigma_G    =    0.1; %
    rho_G      =    10;
    iotat      =    zeros(length(time),1);

    iotat(1)   =    mu_G(1);
    
    deltasimul =    De_t/sum(De_t)*T;     

    nsimul     =    length(deltasimul)+1;   
    NUM(iMC)   =    nsimul;
    diurnal    = [ (0.75 * 1 + 0.25 * exp(-10*1)+0.8);repmat((0.75 * exp(-10*(1:(nsimul-1))/(nsimul-1)) + 0.25 * exp(-10*(1-(1:(nsimul-1))/(nsimul-1)))+0.8)',floor(T*252),1)];%


    rndn1      =  randn(nsimul,1);  
    rndn2      =  rho*rndn1+sqrt(1-rho^2)*randn(nsimul,1); 
  
    
        Xsimul = zeros(nsimul,1);
        Vsimul = zeros(nsimul,1);
        Fsimul = zeros(nsimul,1); 

        Xsimul(1) = X0;    
        Fsimul(1) = F0;
        Vsimul(1)  = diurnal(1)*exp(alpha+beta*Fsimul(1));
            
        Jidx       = binornd(1,lambda*deltabar,nsimul-1,1);
        JumpVectorV=(muJ+sigmaJ*randn(nsimul-1,1)).*Jidx-muJ*lambda*deltasimul;
        JumpVectorX=(mux+sigmax*randn(nsimul-1,1)).*Jidx-mux*lambda*deltasimul;
    
    for i=2:nsimul     % Simulation sample path

        
         Xsimul(i) = Xsimul(i-1)+(mu0+mu1.*Vsimul(i-1))*deltasimul(i-1) + sqrt(Vsimul(i-1))*sqrt(deltasimul(i-1))*rndn1(i)+JumpVectorX(i-1);
         Fsimul(i) = Fsimul(i-1)+kappa*Fsimul(i-1)*deltasimul(i-1)+rndn2(i)*sigma*sqrt(deltasimul(i-1))+JumpVectorV(i-1);
         Vsimul(i) = diurnal(i)*exp(alpha+beta*Fsimul(i));    
         iotat(i)  = iotat(i-1)-rho_G*(iotat(i-1)-mu_G(i-1))*deltasimul(i-1)+sigma_G*sqrt(deltasimul(i-1))*rndn1(i);

    end
    
    if  q >= 1 & iota2true~=0
        
        switch dist
            case 0 % t distribution with nu degrees
                nu  = 7;
                innoveps      = sqrt((nu-2)/nu)*trnd(ones(2*nsimul,1)*nu);
                U = zeros(nsimul,1);
                for jj=(p+q+1):(2*nsimul)
                    U(jj) = innoveps(jj:-1:jj-q)'*[1;thetatrue'];
                end
            case 1 % normal
                U             = simulate(model,2*nsimul);
            case 2 % uniform
                innoveps      = (rand(2*nsimul,1)-0.5)*2*sqrt(3);
                U = zeros(2*nsimul,1);
                for jj=(p+q+1):(2*nsimul)
                    U(jj) = U((jj-1):-1:(jj-p))'*phitrue + innoveps(jj:-1:jj-q)'*[1;thetatrue'];
                end
        end
        
        U = U((nsimul+1):2*nsimul).*sqrt(iota2true).*iotat;
    else
        U         = sqrt(iota2true)*randn(nsimul,1).*iotat;        % Microstructure Noise
    end

    Xtilde        = Xsimul + U;                        % Observation with noise           
    Xtilde        = log(round(exp(Xtilde)*100)/100); %no rounding
    n             = nsimul-1;        
    y             = diff(Xtilde);              % the noisy return
    dX            = diff(Xsimul);              % the non-noisy data
    
    RV_hat(iMC,1)            =     sum(y.^2);   
    RVsub_hat(iMC,1)         =     sum(diff(Xtilde(1:sublag:end)).^2);  
    RV_true(iMC,1)           =     sum(Vsimul(1:end-1).*deltasimul)+sum(JumpVectorX.^2);  

%% ------- Model Selection
%     

    results  = QMLE_MA(y,T,qtrue,quse);
    
    QML_fval(iMC,:) = results.fval;
    QML_var(iMC,:)  = results.iv;
    QML_iota(iMC,:) = results.iota;
    QML_BIC(iMC,1)  = results.BIC;
    QML_AIC(iMC,1)  = results.AIC;
    
    QML_qhat(iMC,:) = results.q;
    QML_ma(:,:,iMC) = results.ma;
    QML_gamma(:,:,iMC)  = results.gamma; 
    QML_rho(:,:,iMC)  = results.rho; 
    
    QML_std(iMC,:) = results.std;
    
    QML_sstd(iMC,:) = results.sstd;
    
    QML_ustd(iMC,:) = results.ustd;


%%
%%  FRK
    
    RK_vec =  zeros(1,length(rkn));  
    for l=1:length(rkn)
        fc =  rkn(l);
        JT =  jkn(l);
        if JT>0
            yjit = diff([mean(Xtilde(1:JT));Xtilde((JT+1):(n-JT));mean(Xtilde(n-JT+1:n))]);
        else
            yjit=y;
        end
        fH =  hkn(l);
        RK_vec(l) = function_FTRK_PZ_multi_all_bandwidths(1,yjit,fH,fc*fH^(-0.6))/T;
    end    
    
    RK_var(iMC,:)  = RK_vec;
    
%% Pre-averaging
  
    PVG_vec = zeros(1,length(knl));  
    PVG_R   = zeros(J+1,length(knl));  
    PVG_r   = zeros(J,length(knl)); 
    PVG_Rstd   = zeros(J+1,length(knl));  
    PVG_rstd   = zeros(J,length(knl)); 
    for l=1:length(knl)
        PVG_results     = PreAvg_JLZ(Xtilde,J,T,knl(l),hnl(l),knpl(l));
        PVG_vec(l)      = PVG_results.Pvg;
        PVG_R(:,l)      = PVG_results.R;
        PVG_r(:,l)      = PVG_results.r(2:end);
        PVG_Rstd(:,l)      = PVG_results.Rstd;
        PVG_rstd(:,l)      = PVG_results.rstd;
   
    end 
    
    PVG_var(iMC,:)      = PVG_vec;
    PVG_gamma(:,:,iMC)  = PVG_R;
    PVG_rho(:,:,iMC)    = PVG_r;
    PVG_gamma_std(:,:,iMC)  = PVG_Rstd;
    PVG_rho_std(:,:,iMC)    = PVG_rstd;

%% 

  t = getCurrentTask();
  took = toc(ticID); 
  
  if isempty(t) % 0 if matlabpool is closed
    fprintf('MC = %5.0f/%5.0f; took %1.3fs\n',iMC,M,took);
  else
    fprintf('MC = %5.0f/%5.0f; worker_id=%2.0f took %1.3fs\n',iMC,M,t.ID,took);
  end

  

end %on iMC loop
toc;
%%
    save(['simu_arma_',num2str(p),'_',num2str(q),'_',num2str(nbseconds),'sec_noise_',num2str((sqrt(iota2true*10^12))),'_',num2str(ii),'_Comparison.mat']);

end
end